<!DOCTYPE html>
<html>

<head>
    <title>OpenAPI power by Index.py</title>
    <meta charset="utf-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1">

    <style>
        * {
            font-family: Menlo, Consolas, "Source Code Pro", Inconsolata, Monaco, "Courier New",
                'Segoe UI', Helvetica, Arial, sans-serif !important;
        }

        h1,
        h2 {
            font-family: 'Segoe UI', Helvetica, Arial, sans-serif !important;
        }

        body {
            margin: 0;
            padding: 0;
        }
    </style>
</head>

<body>
    <redoc spec-url='./json'></redoc>
    <script src="https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js"> </script>
</body>

</html>
